{
   muu Library for TMT Pascal - Autodesk FLI/FLC animation player
   Copyright (C) 1999 Tor Arne Gustad (anden@online.no)

   This source code is licensed under the GNU Public License.
   See the GNU General Public License for more details.
}

UNIT mFLI;

INTERFACE
{$R-,I+,Q-}

USES mFormat,mSurface,Nova,HTimer,Memory,Error;

CONST FLIMagic =        $AF11;
      FLCMagic =        $AF12;
      FrameMagic =      $F1FA;

      Headersize =      128;
      frameHeadersize = 16;
      chunkHeadersize = 6;
      streamingbuffer = 8192;

TYPE FLILoadingMode = (Stream,Preload);

TYPE FLIPaletteT = ARRAY[BYTE,0..2] OF BYTE;

TYPE FLIHeaderT =   RECORD
                      size:     dword;
                      magic:    word;
                      frames:   word;
                      width:    word;
                      height:   word;
                      depth:    word;
                      flags:    word;
                      speed:    word;
                      expand:   ARRAY[0..109] OF byte;
                    END;

TYPE FLIframeHeaderT =  RECORD
                          size:     dword;
                          magic:    word;
                          chunks:   word;
                          expand:   ARRAY[0..7] OF byte;
                        END; 

TYPE FLIchunkHeaderT =  RECORD
                          size:     dword;
                          ctype:    word;
                        END;

TYPE FLCHeaderT =   RECORD
                      size:     dword;
                      magic:    word;
                      frames:   word;
                      width:    word;
                      height:   word;
                      depth:    word;
                      flags:    word;
                      speed:    dword;
                      reserved: word;
                      created:  dword;
                      creator:  dword;
                      updated:  dword;
                      updater:  dword;
                      aspectx:  word;
                      aspecty:  word;
                      reserved2:ARRAY[0..37] OF byte;
                      oframe1:  dword;
                      oframe2:  dword;
                      expand:   ARRAY[0..39] OF byte;
                    END;

TYPE muuFlic =    ^muuFlicT;
     muuFlicT =   OBJECT

                  public
                   CONSTRUCTOR Open(flifile:string;flimode:FLILoadingMode);
                   DESTRUCTOR Close;

                    // read next frame
                   PROCEDURE Frame;
                    // copy frame to user surface
                   PROCEDURE Draw(dx,dy:longint;d_surface:Surface);

                    // user settings
                   PROCEDURE SetFrame(n_frame:dword);
                   PROCEDURE Speed(n_speed:single);     // set speed in msec
                   PROCEDURE Looping(n_loop:boolean);
                   FUNCTION Position:dword;
                   FUNCTION Active: boolean;

                   FUNCTION Width: dword;
                   FUNCTION Height: dword;

                  private

                    // various loaders
                   PROCEDURE Load_FLI(flifile:string);
                   PROCEDURE Load_FLC(flifile:string);

                    // chunk reader code
                   PROCEDURE FLItreatframe;
                   PROCEDURE FLIreadnextframe;

                    // FLC specific decoding
                   PROCEDURE FLItreatchunkCOLOR256(chunk:pointer;VAR destpal:FLIPaletteT);
                   PROCEDURE FLItreatchunkSS2(chunk:pointer;dest:pointer);

                    // FLI/FLC decoding
                   PROCEDURE FLItreatchunkCOLOR(chunk:pointer;VAR destpal:FLIPaletteT);
                   PROCEDURE FLItreatchunkLC(chunk:pointer;dest:pointer);
                   PROCEDURE FLItreatchunkBLACK(dest:pointer);
                   PROCEDURE FLItreatchunkBRUN(chunk:pointer;dest:pointer);
                   PROCEDURE FLItreatchunkCOPY(chunk:pointer;dest:pointer);

                   // fli header structures
                   FLIframeheader:  FLIframeHeaderT;
                   FLIchunkheader:  FLIchunkHeaderT;

                    // data blocks
                   m_flidata:       pointer;
                   m_framedata:     pointer;
                   m_chunkdata:     pointer;
                   m_palette:       FLIPaletteT;

                    // data positions
                   m_pos:           dword;
                   m_maxpos:        dword;
                   m_frame:         dword;
                   m_file:          file;

                    // target surface
                   buffer:          Surface;
                    
                    // common information
                   m_frames:        dword;
                   m_firstframe:    dword;
                   m_loopframe:     dword;

                    // settings
                   m_speed:         single;
                   m_active:        boolean;
                   m_looping:       boolean;
                   m_mode:          FLILoadingMode;
                    
                    // timer data
                   m_timer:         Timer;
                 END;
                   

IMPLEMENTATION

CONSTRUCTOR muuFlicT.Open;
  VAR ch:   FLIHeaderT;
BEGIN
    // open flic and get size
  system.assign(m_file,flifile);
  {$I-} system.reset(m_file,1); {$I+}
  IF (system.IOresult <> 0) THEN 
    fail('muuFlic: FLIC file ['+flifile+'] not found');
  m_maxpos := filesize(m_file);
  m_pos := 0;

  m_mode := flimode;

  IF (m_maxpos <= Headersize) THEN 
    fail('muuFlic: FLIC file ['+flifile+'] is too small');

  blockread(m_file,ch,Headersize);
  system.close(m_file);

  CASE (ch.magic) OF
    FLIMagic:   Load_FLI(flifile);
    FLCMagic:   Load_FLC(flifile);
   
   ELSE
     fail('muuFlic: ['+flifile+'] is not a FLIC file');

  END;

    // set starting variables
  m_active := true;
  m_looping := true;
  m_frame := 0;
  m_pos := m_firstframe;

    // allocate and start timer
  new(m_timer,Create);
  m_timer^.Start;
END;

DESTRUCTOR muuFlicT.Close;
BEGIN
  IF (m_flidata = NIL) THEN 
    exit;

    // free data
  IF (m_mode = Preload) THEN
    freemem(m_flidata,m_maxpos+128)
  ELSE BEGIN
    freemem(m_flidata,streamingbuffer);
    system.close(m_file);
  END;
    
  freemem(m_framedata,buffer^.Width*buffer^.Height*2);
  freemem(m_chunkdata,buffer^.Width*buffer^.Height*2);
 
  dispose(buffer,Close);
  dispose(m_timer,Destroy);
END;

PROCEDURE muuFlicT.Load_FLI;
  VAR Header:   FLIHeaderT;
BEGIN
    // open flic and get size
  system.assign(m_file,flifile);
  {$I-} system.reset(m_file,1); {$I+}
  m_maxpos := filesize(m_file);
  m_pos := 0;

  IF (m_mode = Preload) THEN
  BEGIN
    // allocate memory for data
    getmem(m_flidata,m_maxpos+Headersize);

    // read flic data
    blockread(m_file,m_flidata^,m_maxpos);
    system.close(m_file);
  END ELSE BEGIN
    getmem(m_flidata,streamingbuffer);
    blockread(m_file,m_flidata^,Headersize);
  END;

    // get header data
  memcpy(ptr(ofs(m_flidata^)+m_pos),ptr(ofs(Header)),Headersize);

    // get frame/chunk data size
  getmem(m_framedata,Header.Width*Header.Height*2);
  getmem(m_chunkdata,Header.Width*Header.Height*2);

  m_firstframe := Headersize;
  m_loopframe := Headersize;

    // allocate buffer surface
  buffer := Surface(Header.width,Header.height,Format(VPal));

    // set starting variables
  m_active := true;
  m_looping := true;

    // FLI speed is given in 1/70 second, make it 1/1000 second
  m_speed := (Header.Speed/70)*1000;

  m_frame := 0;
  m_frames := Header.frames;
END;

PROCEDURE muuFlicT.Load_FLC;
  VAR Header:   FLCHeaderT;
BEGIN
    // open flic and get size
  system.assign(m_file,flifile);
  {$I-} system.reset(m_file,1); {$I+}
  m_maxpos := filesize(m_file);
  m_pos := 0;

  IF (m_mode = Preload) THEN
  BEGIN
    // allocate memory for data
    getmem(m_flidata,m_maxpos+Headersize);

    // read flic data
    blockread(m_file,m_flidata^,m_maxpos);
    system.close(m_file);
  END ELSE BEGIN
    getmem(m_flidata,streamingbuffer);
    blockread(m_file,m_flidata^,Headersize);
  END;

  memcpy(ptr(ofs(m_flidata^)+m_pos),ptr(ofs(Header)),Headersize);

    // get frame/chunk data size
  getmem(m_framedata,Header.Width*Header.Height*2);
  getmem(m_chunkdata,Header.Width*Header.Height*2);

    // go directly to first frame data
  m_firstframe := Header.oframe1;
  m_loopframe := Header.oframe2;

    // allocate buffer surface
  buffer := Surface(Header.width,Header.height,Format(VPal));

    // set starting variables
  m_active := true;
  m_looping := true;

    // speed is 1/1000 second
  m_speed := Header.Speed;

  m_frame := 0;
  m_frames := Header.frames;
END;

PROCEDURE muuFlicT.Frame;
  VAR time,otime:   single := 0;
BEGIN
  IF NOT m_active THEN exit;

  IF (m_speed > m_timer^.Elapsed) THEN
    Exit;

  m_timer^.Reset;

  FLIreadnextframe;
  FLItreatframe;
END;

PROCEDURE muuFlicT.Draw;
  VAR c: byte := 0;
BEGIN
    // copy palette data to buffer surface (VPal)
  FOR c := 0 TO 255 DO
  BEGIN
    buffer^.sFormat.Palette[c,VRed] := m_palette[c,0];
    buffer^.sFormat.Palette[c,VGreen] := m_palette[c,1];
    buffer^.sFormat.Palette[c,VBlue] := m_palette[c,2];
  END;
    
    // copy surface data
  d_surface^.Copy(dx,dy,buffer,Area(buffer));
END;


PROCEDURE muuFlicT.SetFrame;
  VAR c: dword := 0;
BEGIN
  IF (n_frame > m_frames) THEN 
    n_frame := m_frames;

  IF (n_frame = 0) THEN 
    n_frame := 1;

  m_pos := m_firstframe;

  FOR c := 1 TO n_frame DO
  BEGIN
    FLIreadnextframe;
    FLItreatframe;
  END;
  m_frame := n_frame;
END;

PROCEDURE muuFlicT.Speed;
  m_speed := n_speed;

PROCEDURE muuFlicT.Looping;
  m_looping := n_loop;

FUNCTION muuFlicT.Position;
  result := m_frame;

FUNCTION muuFlicT.Active;
  result := m_active;

FUNCTION muuFlicT.Width;
  result := buffer^.Width;

FUNCTION muuFlicT.Height;
  result := buffer^.Height;

PROCEDURE muuFlicT.FLItreatchunkCOLOR256; assembler;
ASM
  mov  esi,   chunk
  mov  eax,   [ebp] + offset destpal
  mov  edi,   eax

  xor  ebx,   ebx
  mov  bx,    [esi]    // number of packets
  add  esi,   2

  xor  eax,   eax
  xor  ecx,   ecx

@@packetloop:
  mov  al,    [esi]    // colors to skip
  mov  cl,    [esi+1]  // colors to change
  add  esi,   2

  mov  edx,   eax      // multiply EAX with 3
  shl  edx,   1        // ((x*2)+x) = (x*3)
  add  eax,   edx      // ...using EDX as storage
  add  edi,   eax      // then skip the colors

  cmp  ecx,   0        // if ECX is 0, then 256 colors is to be changed
  sete ch              // if it is, we set bit CH to 1, which makes (E)CX 256
  mov  edx,   ecx      //
  shl  edx,   1        // multiply ECX by 3, too
  add  ecx,   edx      //

  rep  movsb

  dec  ebx
  jnz  @@packetloop
END;

PROCEDURE muuFlicT.FLItreatchunkSS2; assembler;
  VAR scrtab:   pointer := buffer^.hmultab;
      scrwidth: dword := buffer^.Width;
ASM
  mov  esi,  chunk
  mov  edi,  dest

    // get #lines in chunk
  mov  dx,   [esi]
  add  esi,  2

@@lineloop:
  push edi
  push edx

  xor  eax,  eax
  xor  ebx,  ebx
  xor  ecx,  ecx

    // get packet information
  mov  bx,   [esi]
  add  esi,  2

  mov  ax,   bx
  shr  ax,   14
  
    // skip lines?
  cmp  al,   3
  je   @@lineskip

    // last pixel in line?
  cmp  al,   2
  je   @@lastpixel

  and  ebx,  $3FFF

    // third pitfall
  cmp  ebx,  0
  je   @@endline

@@packetloop:
  xor  eax,  eax

  mov  al,   [esi]          // bytes to skip in line
  add  edi,  eax
  mov  cl,   [esi+1]        // bytes to change in line
  add  esi,  2

    // second pitfall
  cmp  cl,   0
  je   @@packetloop

    // first pitfall
  or   cl,   cl
  jns  @@datacount          // if CL>0, then comes CL bytes of data to copy
                            // if CL<0, one color shall be put -CL times
  neg  cl
  mov  ax,   [esi]
  add  esi,  2
  rep  stosw
  jmp  @@endpacket

@@datacount:
  rep  movsw
  jmp  @@endpacket

@@endpacket:
  dec  ebx
  jnz  @@packetloop

@@endline:
  pop  edx
  pop  edi
  add  edi,  scrwidth
  dec  edx
  jnz  @@lineloop

  jmp  @@endchunk

    // skip -bx number of lines
@@lineskip:
  pop  edx
  pop  edi
  neg  bx
  mov  eax,  scrtab
  mov  ebx,  [eax+ebx*4]
  add  edi,  ebx
  jmp  @@lineloop

    // put bl as last pixel on current line
@@lastpixel:
  add  edi,  scrwidth
  dec  edi
  mov  [edi],  bl
  pop  edx
  pop  edi
  jmp  @@lineloop

@@endchunk:
END;


PROCEDURE muuFlicT.FLItreatchunkCOLOR; assembler;
ASM
  mov  esi,   chunk
  mov  eax,   [ebp] + offset destpal
  mov  edi,   eax

  xor  ebx,   ebx
  mov  bx,    [esi]    // number of packets
  add  esi,   2

  xor  eax,   eax
  xor  ecx,   ecx

@@packetloop:
  mov  al,    [esi]    // colors to skip
  mov  cl,    [esi+1]  // colors to change
  add  esi,   2

  mov  edx,   eax      // multiply EAX with 3
  shl  edx,   1        // ((x*2)+x) = (x*3)
  add  eax,   edx      // ...using EDX as storage
  add  edi,   eax      // then skip the colors

  cmp  ecx,   0        // if ECX is 0, then 256 colors is to be changed
  sete ch              // if it is, we set bit CH to 1, which makes (E)CX 256
  mov  edx,   ecx      //
  shl  edx,   1        // multiply ECX by 3, too
  add  ecx,   edx      //

  rep  movsb

  dec  ebx
  jnz  @@packetloop

    // change to 8bit palette
  mov  eax,   [ebp] + offset destpal
  mov  esi,   eax
  mov  ecx,   767
@@upgrade:
  mov  al,    [esi]
  shl  al,    2
  mov  [esi], al
  inc  esi
  dec  ecx
  jnz  @@upgrade
END;

PROCEDURE muuFlicT.FLItreatchunkLC; assembler;
  VAR scrtab:   pointer := buffer^.hmultab;
      scrwidth: dword := buffer^.Width;
ASM
  mov  esi,  chunk
  mov  edi,  dest
  mov  eax,  scrtab

  xor  ebx,  ebx
  mov  bx,   [esi]          // lines to skip (*Width)
  mov  ebx,  [eax+ebx*4]
  add  edi,  ebx

  xor  ebx,  ebx
  mov  bx,   [esi+2]        // lines to change
  add  esi,  4

  xor  eax,  eax
  xor  ecx,  ecx
  xor  edx,  edx

@@lineloop:
  mov  dl,   [esi]          // packets in line
  inc  esi
  push edi                  // preserve EDI for an easier jump to next line
  cmp  edx,  0              // has the line really changed?
  je   @@endline

@@packetloop:
  mov  al,   [esi]          // bytes to skip in line
  add  edi,  eax
  mov  cl,   [esi+1]        // bytes to change in line
  add  esi,  2

    // first pitfall
  or   cl,   cl
  jns  @@datacount          // if CL>0, then comes CL bytes of data to copy
                            // if CL<0, one color shall be put -CL times
  neg  cl
  mov  al,   [esi]
  inc  esi
  rep  stosb
  jmp  @@endpacket

@@datacount:
  rep  movsb
  jmp  @@endpacket

@@endpacket:
  dec  edx
  jnz  @@packetloop

@@endline:
  pop  edi
  add  edi,  scrwidth
  dec  ebx
  jnz  @@lineloop
END;

PROCEDURE muuFlicT.FLItreatchunkBLACK;
  memset(dest,buffer^.Size,$00);

PROCEDURE muuFlicT.FLItreatchunkBRUN; assembler;
  VAR lines:    dword := buffer^.Height;
      scrwidth: dword := buffer^.Width;
ASM
  mov  esi,  chunk
  mov  edi,  dest
  mov  ebx,  lines          // all lines shall be put

  xor  eax,  eax
  xor  ecx,  ecx
  xor  edx,  edx

@@lineloop:
  mov  dl,   [esi]          // #packets in line
  inc  esi
  mov  edx,  scrwidth       // ignore #packets count, use max count instead
  xor  ecx,  ecx 
  push edi                  // preserve EDI

@@packetloop:
  mov  cl,   [esi]          // colors to put
  or   cl,   cl             // if CL<0, -CL bytes of data is to come
  js   @@datacount          // if CL>0, one byte shall be put CL times

  sub  edx,  ecx            // decrease "pixels left"
  mov  al,   [esi+1]        // color to put CL times
  add  esi,  2
  rep  stosb
  jmp  @@endpacket

@@datacount:
  neg  cl
  sub  edx,  ecx            // decrease "pixels left"
  inc  esi
  rep  movsb
  jmp  @@endpacket

@@endpacket:
  cmp  edx,  0
  ja   @@packetloop // continue until line finished
                    
@@endline:
  pop  edi
  add  edi,  scrwidth
  dec  ebx
  jnz  @@lineloop
END;

PROCEDURE muuFlicT.FLItreatchunkCOPY;
  memcpy(chunk,dest,buffer^.Size);

PROCEDURE muuFlicT.FLItreatframe;
  VAR c,framepos: dword := 0;
BEGIN
  FOR c := 1 TO FLIframeheader.chunks DO
  BEGIN
        // get chunk header
    memcpy(ptr(ofs(m_framedata^)+framepos),ptr(ofs(FLIchunkheader)),chunkHeadersize);
    framepos +:= chunkHeadersize;
        // get chunk data
    memcpy(ptr(ofs(m_framedata^)+framepos),m_chunkdata,FLIchunkheader.size-chunkHeadersize);
    framepos +:= (FLIchunkheader.size-chunkHeadersize);

    CASE FLIchunkheader.ctype OF
       4: FLItreatchunkCOLOR256(m_chunkdata,m_palette);
       7: FLItreatchunkSS2(m_chunkdata,buffer^.Data);
      11: FLItreatchunkCOLOR(m_chunkdata,m_palette);
      12: FLItreatchunkLC(m_chunkdata,buffer^.Data);
      13: FLItreatchunkBLACK(buffer^.Data);
      15: FLItreatchunkBRUN(m_chunkdata,buffer^.Data);
      16: FLItreatchunkCOPY(m_chunkdata,buffer^.Data);
      18: BEGIN END; //skip "postage stamp" frames (preview frames)

        ELSE
          fail('muuFlic: Unsupported chunk type');
     END;
  END;
END;

PROCEDURE muuFlicT.FLIreadnextframe;
BEGIN
  IF (m_pos >= m_maxpos) THEN
  BEGIN
    IF m_looping THEN
    BEGIN
      m_pos := m_loopframe;
      m_frame := 0;
    END;
    IF NOT m_looping THEN
    BEGIN
      m_active := false;
      exit;
    END;
  END;

  IF (m_mode = Preload) THEN
  BEGIN

    // get frame header
    memcpy(ptr(ofs(m_flidata^)+m_pos),ptr(ofs(FLIframeheader)),frameHeadersize);
    m_pos +:= frameHeadersize;

    IF (FLIframeheader.magic <> FrameMagic) THEN 
      fail('muuFlic: FLIC frame header error');

    // get frame data
    memcpy(ptr(ofs(m_flidata^)+m_pos),m_framedata,FLIframeheader.size-frameHeadersize);
  END ELSE BEGIN

    // read frame header
    seek(m_file,m_pos);
    blockread(m_file,FLIframeheader,frameHeadersize);
    m_pos +:= frameHeadersize;

    IF (FLIframeheader.magic <> FrameMagic) THEN 
      fail('muuFlic: FLIC frame header error');

    // read frame data
    blockread(m_file,m_framedata^,FLIframeheader.size-frameheadersize);
  END;

  m_pos +:= FLIframeheader.size-frameHeadersize;
  m_frame +:= 1;
END;


end.
